# ========================================================================================================
#
# John Pritchard
#
# ========================================================================================================
try:
  import numpy
  from pipeline_display import *
  from pipeline_product import *
  from reflex_plot_widgets import *
  from numpy.polynomial import Polynomial
  numpy.seterr(invalid='ignore')
  import giraf_der_snr
  from giSpectrumDisplay import *
  #from pipeline_display import SpectrumDisplay as giSpectrumDisplay
  
  from astropy.coordinates import SkyCoord
  from astropy.table import Table
  #from astroquery.utils import TableList
  import astropy.units as u
except ImportError:
  donothing=1

def get_attrs(klass):
  return [k for k in list(klass.__dict__.keys())
            if not k.startswith('__')
            and not k.endswith('__')]

## Potentially interesting code here...
## http://python4esac.github.io/plotting/specnorm.html

## These are the GIRAFFE routines, hacked from the FORS ones,
## see below for the remaining FORS ones as examples...
class PlotableFlat(object) :
  def __init__(self, fits_flat):
    self.flat      = PipelineProduct(fits_flat)
    self.flatdisp  = ImageDisplay()
    self.loadFromFits()

  def loadFromFits(self) :
    #Reading the flat image
    self.flat.readImage()

  def plot(self, subplot, title, tooltip):
    self.flatdisp.setLabels('X [pix]', 'Y [pix]')
    self.flatdisp.display(subplot, title, tooltip, self.flat.image)

################################################################################

class PlotableFlatWithLOCcentroid(object) :
  def __init__(self, fits_flat, fits_ff_loccentroid):
    self.Xsectplot = -1
    self.flat      = PipelineProduct(fits_flat)
    self.ff_loccentroid = None
    if fits_ff_loccentroid is not None:
      self.ff_loccentroid = PipelineProduct(fits_ff_loccentroid)
    self.flatdisp  = ImageDisplay()
    self.Xsectplot = giSpectrumDisplay()
    self.loadFromFits()

    self.NAXIS1=self.flat.all_hdu[0].header['NAXIS1']
    self.NAXIS2=self.flat.all_hdu[0].header['NAXIS2']
    # Params are available via the interactive_app.getCurrentParameterValue()
    # but I can't figure out how to access them from here... so I need to
    # make my own param dictionary...
    i=0
    pFound=True
    self.params = {}
    self.params[self.flat.all_hdu[0].header['HIERARCH ESO PRO REC1 PARAM%d NAME' %(i+1)]] = self.flat.all_hdu[0].header['HIERARCH ESO PRO REC1 PARAM%d VALUE' %(i+1)]
    while pFound :
        try :
            self.params[self.flat.all_hdu[0].header['HIERARCH ESO PRO REC1 PARAM%d NAME' %(i+1)]] = self.flat.all_hdu[0].header['HIERARCH ESO PRO REC1 PARAM%d VALUE' %(i+1)]
            i+=1
        except :
            pFound=False
    self.sloc_start=int(self.params['sloc-start'])
    self.setXsectRow(self.sloc_start)

  def loadFromFits(self) :
    #Reading the flat image
    self.flat.readImage()
    if self.ff_loccentroid is not None:
      self.ff_loccentroid.readImage()
      self.fibres = self.ff_loccentroid.image.shape[1]

      #Creating the points to plot based on the polynomial traces
      self.ypos_fibres = list(range(self.ff_loccentroid.image.shape[0]))
      self.xpos_fibres = []
      for fibre in numpy.arange(self.fibres) :
        self.xpos_fibres.append(self.ff_loccentroid.all_hdu[0].data[:,fibre]+1.)


  def plot(self, subplot, title, tooltip):
    subplot.cla()
    self.flatdisp.setLabels('X [pix]', 'Y [pix]')
    self.flatdisp.display(subplot, title, tooltip, self.flat.image)
    lss=self.sloc_start
    if lss == -1 :
        lss=int(self.NAXIS2/2)
    if self.ff_loccentroid is not None:
      subplot.autoscale(enable=False)
      ## EVEN number subslits...
      for fibre in numpy.arange(self.fibres)[numpy.where(self.ff_loccentroid.all_hdu[1].data['SSN']%2 == 0)] :
        subplot.plot(
          self.xpos_fibres[fibre], self.ypos_fibres,
          linestyle='solid',color='red'
        )
      ## ODD number subslits...
      for fibre in numpy.arange(self.fibres)[numpy.where(self.ff_loccentroid.all_hdu[1].data['SSN']%2 == 1)] :
        subplot.plot(
          self.xpos_fibres[fibre], self.ypos_fibres,
          linestyle='solid',color='yellow'
        )
      ## Label the subslits...
      for SS in numpy.unique(self.ff_loccentroid.all_hdu[1].data['SSN']):
        subplot.annotate(
          str(SS),
          xy=(numpy.mean(numpy.array(self.xpos_fibres)[numpy.where(self.ff_loccentroid.all_hdu[1].data['SSN'] == SS)][:,lss]),self.NAXIS2),
          xycoords='data', horizontalalignment='center', verticalalignment='top'
        )
      ## Label the fibres...
      for i in numpy.unique(self.ff_loccentroid.all_hdu[1].data['FPS']):
        subplot.annotate(
          str(i),
          xy=(numpy.array(self.xpos_fibres)[numpy.where(self.ff_loccentroid.all_hdu[1].data['FPS'] == i)][0][lss],lss),
          xycoords='data', horizontalalignment='center', verticalalignment='bottom', rotation=90.
        )
    ## A horizontal line at the sloc-start value...
    subplot.plot([0,self.NAXIS1], [lss,lss], linestyle='solid', color='black', linewidth=2)

  def setXsectRow(self, row):
      self.XsectRow=row
      if row == -1 :
          self.XsectRow=int(self.NAXIS2/2)

  def plotXsection(self, subplot, title, tooltip):
    subplot.cla()
    self.Xsectplot.setLabels('X [pix]', 'Y [ADU]')
    self.Xsectplot.display(subplot, title, tooltip, numpy.arange(self.NAXIS1)+1, self.flat.image[self.XsectRow,:])
    SSylab=subplot.get_ylim()[1]*0.99
    lss=self.sloc_start
    if lss == -1 :
        lss=int(self.NAXIS2/2)
    if self.ff_loccentroid is not None:
      subplot.autoscale(enable=False)
      ## EVEN number subslits...
      for fibre in numpy.arange(self.fibres)[numpy.where(self.ff_loccentroid.all_hdu[1].data['SSN']%2 == 0)] :
        subplot.axvline(
          self.xpos_fibres[fibre][self.XsectRow],
          linestyle='solid',color='red'
        )
      ## ODD number subslits...
      for fibre in numpy.arange(self.fibres)[numpy.where(self.ff_loccentroid.all_hdu[1].data['SSN']%2 == 1)] :
        subplot.axvline(
          self.xpos_fibres[fibre][self.XsectRow],
          linestyle='solid',color='yellow'
        )
      ## Label the subslits...
      for SS in numpy.unique(self.ff_loccentroid.all_hdu[1].data['SSN']):
        subplot.annotate(
          str(SS),
          xy=(numpy.mean(numpy.array(self.xpos_fibres)[numpy.where(self.ff_loccentroid.all_hdu[1].data['SSN'] == SS)][:,self.XsectRow]),SSylab),
          xycoords='data', horizontalalignment='center', verticalalignment='top'
        )
      ## Label the fibres...
      for i in numpy.unique(self.ff_loccentroid.all_hdu[1].data['FPS']):
        subplot.annotate(
          str(i),
          xy=(numpy.array(self.xpos_fibres)[numpy.where(self.ff_loccentroid.all_hdu[1].data['FPS'] == i)][0][self.XsectRow],0),
          xycoords='data', horizontalalignment='center', verticalalignment='bottom', rotation=90.
        )

################################################################################

class PlotableFlatWithPSFcentroid(object) :
  def __init__(self, fits_flat, fits_ff_psfcentroid):
    self.flat      = PipelineProduct(fits_flat)
    self.ff_psfcentroid = None
    if fits_ff_psfcentroid is not None:
      self.ff_psfcentroid = PipelineProduct(fits_ff_psfcentroid)
    self.flatdisp  = ImageDisplay()
    self.Xsectplot = giSpectrumDisplay()
    self.loadFromFits()

    self.NAXIS1=self.flat.all_hdu[0].header['NAXIS1']
    self.NAXIS2=self.flat.all_hdu[0].header['NAXIS2']
    # Params are available via the interactive_app.getCurrentParameterValue()
    # but I can't figure out how to access them from here... so I need to
    # make my own param dictionary...
    i=0
    pFound=True
    self.params = {}
    self.params[self.flat.all_hdu[0].header['HIERARCH ESO PRO REC1 PARAM%d NAME' %(i+1)]] = self.flat.all_hdu[0].header['HIERARCH ESO PRO REC1 PARAM%d VALUE' %(i+1)]
    while pFound :
        try :
            self.params[self.flat.all_hdu[0].header['HIERARCH ESO PRO REC1 PARAM%d NAME' %(i+1)]] = self.flat.all_hdu[0].header['HIERARCH ESO PRO REC1 PARAM%d VALUE' %(i+1)]
            i+=1
        except :
            pFound=False
    self.sloc_start=int(self.params['sloc-start'])
    self.setXsectRow(self.sloc_start)

  def loadFromFits(self) :
    #Reading the flat image
    self.flat.readImage()
    if self.ff_psfcentroid is not None:
      self.ff_psfcentroid.readImage()
      self.fibres = self.ff_psfcentroid.image.shape[1]

      #Creating the points to plot based on the polynomial traces
      self.ypos_fibres = list(range(self.ff_psfcentroid.image.shape[0]))
      self.xpos_fibres = []
      for fibre in range(self.fibres) :
        self.xpos_fibres.append(self.ff_psfcentroid.all_hdu[0].data[:,fibre]+1.)


  def plot(self, subplot, title, tooltip):
    subplot.cla()
    self.flatdisp.setLabels('X [pix]', 'Y [pix]')
    self.flatdisp.display(subplot, title, tooltip, self.flat.image)
    lss=self.sloc_start
    if lss == -1 :
        lss=int(self.NAXIS2/2)
    if self.ff_psfcentroid is not None:
      subplot.autoscale(enable=False)
      ## EVEN number subslits...
      for fibre in numpy.arange(self.fibres)[numpy.where(self.ff_psfcentroid.all_hdu[1].data['SSN']%2 == 0)] :
        subplot.plot(
          self.xpos_fibres[fibre], self.ypos_fibres,
          linestyle='solid',color='blue'
        )
      ## ODD number subslits...
      for fibre in numpy.arange(self.fibres)[numpy.where(self.ff_psfcentroid.all_hdu[1].data['SSN']%2 == 1)] :
        subplot.plot(
          self.xpos_fibres[fibre], self.ypos_fibres,
          linestyle='solid',color='green'
        )
      ## Label the subslits...
      for SS in numpy.unique(self.ff_psfcentroid.all_hdu[1].data['SSN']):
          subplot.annotate(
            str(SS),
            xy=(numpy.mean(numpy.array(self.xpos_fibres)[numpy.where(self.ff_psfcentroid.all_hdu[1].data['SSN'] == SS)][:,lss]),self.NAXIS2),
            xycoords='data', horizontalalignment='center', verticalalignment='top'
          )
      ## Label the fibres...
      for i in numpy.unique(self.ff_psfcentroid.all_hdu[1].data['FPS']):
        subplot.annotate(
          str(i),
          xy=(numpy.array(self.xpos_fibres)[numpy.where(self.ff_psfcentroid.all_hdu[1].data['FPS'] == i)][0][lss],lss),
          xycoords='data', horizontalalignment='center', verticalalignment='bottom', rotation=90.
        )
    ## A horizontal line at the sloc-start value...
    subplot.plot([0,self.NAXIS1], [lss,lss],
                 linestyle='solid', color='black', linewidth=2)

  def setXsectRow(self, row):
      self.XsectRow=row
      if row == -1 :
          self.XsectRow=int(self.NAXIS2/2)

  def plotXsection(self, subplot, title, tooltip):
    subplot.cla()
    self.Xsectplot.setLabels('X [pix]', 'Y [ADU]')
    self.Xsectplot.display(subplot, title, tooltip, numpy.arange(self.NAXIS1)+1, self.flat.image[self.XsectRow,:])
    SSylab=subplot.get_ylim()[1]*0.99
    lss=self.sloc_start
    if lss == -1 :
        lss=int(self.NAXIS2/2)
    if self.ff_psfcentroid is not None:
      subplot.autoscale(enable=False)
      ## EVEN number subslits...
      for fibre in numpy.arange(self.fibres)[numpy.where(self.ff_psfcentroid.all_hdu[1].data['SSN']%2 == 0)] :
        subplot.axvline(
          self.xpos_fibres[fibre][self.XsectRow],
          linestyle='solid',color='blue'
        )
      ## ODD number subslits...
      for fibre in numpy.arange(self.fibres)[numpy.where(self.ff_psfcentroid.all_hdu[1].data['SSN']%2 == 1)] :
        subplot.axvline(
          self.xpos_fibres[fibre][self.XsectRow],
          linestyle='solid',color='green'
        )
      ## Label the subslits...
      for SS in numpy.unique(self.ff_psfcentroid.all_hdu[1].data['SSN']):
        subplot.annotate(
          str(SS),
          xy=(numpy.mean(numpy.array(self.xpos_fibres)[numpy.where(self.ff_psfcentroid.all_hdu[1].data['SSN'] == SS)][:,self.XsectRow]),SSylab),
          xycoords='data', horizontalalignment='center', verticalalignment='top'
        )
      ## Label the fibres...
      for i in numpy.unique(self.ff_psfcentroid.all_hdu[1].data['FPS']):
        subplot.annotate(
          str(i),
          xy=(numpy.array(self.xpos_fibres)[numpy.where(self.ff_psfcentroid.all_hdu[1].data['FPS'] == i)][0][self.XsectRow],0),
          xycoords='data', horizontalalignment='center', verticalalignment='bottom', rotation=90.
        )


################################################################################

class PlotableARC_RBNSPECTRA(object) :
  def __init__(self, fits_arc_rbnspectra):
    self.rbnspectra      = PipelineProduct(fits_arc_rbnspectra)
    self.rbnspectradisp  = ImageDisplay()
    self.loadFromFits()

  def loadFromFits(self) :
    #Reading the flat image
    self.rbnspectra.readImage()

  def plot(self, subplot, title, tooltip):
    self.rbnspectradisp.setLabels('Fibre', 'Y [pix]')
    self.rbnspectradisp.display(subplot, title, tooltip, self.rbnspectra.image)



################################################################################

class PlotableSCI_RBNSPECTRA(object) :
  def __init__(self, fits_sci_rbnspectra):
    self.rbnspectra      = PipelineProduct(fits_sci_rbnspectra)
    self.rbnspectradisp  = ImageDisplay()
    self.loadFromFits()

  def loadFromFits(self) :
    #Reading the flat image
    self.rbnspectra.readImage()

  def plot(self, subplot, title, tooltip):
    self.rbnspectradisp.setLabels('Fibre', 'Y [pix]')
    self.rbnspectradisp.display(subplot, title, tooltip, self.rbnspectra.image)

  def getObjectInPosition(self, xpos) :
    return int(xpos)

################################################################################

class PlotableExtractedScience :
    def __init__(self, fits_extractedscience, fits_extractederrors, tblLists, fib_id=-1):
        self.fib_id = fib_id
        self.extractedscience  = {}
        self.extractederrors   = {}
        self.IDPs  = {}
        self.tblLists = tblLists
        self.FIBER_SETUPs = {}
        self.PLparams = {}
        for f in fits_extractedscience :
            HDUlist=pyfits.open(f.name)
            CRCM=HDUlist[0].header.get('CMETHOD','NONE')
            self.INS_MODE=HDUlist[0].header['HIERARCH ESO INS MODE']
            self.RADEC=SkyCoord(HDUlist[0].header['RA'],HDUlist[0].header['DEC'], unit=(u.deg, u.deg), frame="icrs")
            self.extractedscience[CRCM] = PipelineProduct(f)
            i=0
            plp={}
            while "HIERARCH ESO PRO REC1 PARAM%d NAME" %(i+1) in self.extractedscience[CRCM].all_hdu[0].header :
                plp[self.extractedscience[CRCM].all_hdu[0].header["HIERARCH ESO PRO REC1 PARAM%d NAME" %(i+1)]] = self.extractedscience[CRCM].all_hdu[0].header["HIERARCH ESO PRO REC1 PARAM%d VALUE" %(i+1)]
                i+=1
            self.PLparams[CRCM]=plp
            HDUlist.close()
            self.FIBER_SETUPs[CRCM]=Table.read(f.name)
            self.FIBER_SETUPs[CRCM]["RA"].unit=u.deg
            self.FIBER_SETUPs[CRCM]["DEC"].unit=u.deg
        for f in fits_extractederrors :
            HDUlist=pyfits.open(f.name)
            CRCM=HDUlist[0].header.get('CMETHOD','NONE')
            self.extractederrors[CRCM]  = PipelineProduct(f)
            HDUlist.close()
        self.spectrumdisplay   = giSpectrumDisplay()
        self.loadFromFits()

        self.sky=None
        self.skyApplyTo = None
        self.skyMethod = None
        self.skyStatistic = None
        self.skyNumFibs = None
        self.skyFibIsSelected = None
        self.skyRejected = None
        #self.computeSky("All", "All Sky Fibres", "median" )
        #self.computeSky("All", "Best <N> Fibres by S/N", "mean" )
        #self.displaySky = True
        #self.displaySignal = True
        #self.displaySignalMinusSky = True

        self.NoCRC_displaySignal = True
        self.NoCRC_displaySignalMinusSky = True
        self.NoCRC_displaySky = False
        self.NoCRC_displayErr = False
        self.AS_displaySignal = True
        self.AS_displaySignalMinusSky = True
        self.AS_displaySky = False
        self.AS_displayErr = False
        self.PC_displaySignal = True
        self.PC_displaySignalMinusSky = True
        self.PC_displaySky = False
        self.PC_displayErr = False
        self.MED_displaySignal = True
        self.MED_displaySignalMinusSky = True
        self.MED_displaySky = False
        self.MED_displayErr = False

    def loadFromFits(self) :
        self.nobj   = {}
        self.crpix  = {}
        self.crval  = {}
        self.cdelt  = {}
        self.bunit  = {}
        self.nwave  = {}
        self.isARG  = {}
        self.wave   = {}
        self.wave   = {}
        for CRCM in list(self.extractederrors.keys()) :
            self.extractederrors[CRCM].readImage()
        for CRCM in list(self.extractedscience.keys()) :
            self.extractedscience[CRCM].readImage()
            self.nobj[CRCM]   = self.extractedscience[CRCM].image.shape[1]
            self.crpix[CRCM]  = self.extractedscience[CRCM].readKeyword('CRPIX2', 0)
            self.crval[CRCM]  = self.extractedscience[CRCM].readKeyword('CRVAL2', 0)
            self.cdelt[CRCM]  = self.extractedscience[CRCM].readKeyword('CDELT2', 0)
            self.bunit[CRCM]  = self.extractedscience[CRCM].readKeyword('BUNIT', 0)
            self.nwave[CRCM]  = self.extractedscience[CRCM].image.shape[0]
            self.isARG[CRCM]  = self.extractedscience[CRCM].readKeyword('HIERARCH ESO INS MODE', 0) == 'ARG'
            self.wave[CRCM]   = (numpy.arange(1, self.nwave[CRCM]+1, 1) - self.crpix[CRCM]) * self.cdelt[CRCM] + self.crval[CRCM]
        self.setSNRarr()
        self.init_object_data()
        self.selectHighestSNR()
        self.setFluxSelected()

    def init_object_data(self, CRCM='NONE'):
        self.FIBER_SETUPs[CRCM]["SNR"]=self.SNRarr[CRCM]
        if self.isARG[CRCM] :
            self.objectList = self.FIBER_SETUPs[CRCM]['Retractor']
        else :
            self.objectList = self.FIBER_SETUPs[CRCM]['OBJECT']
        self.FPSList = self.FIBER_SETUPs[CRCM]['FPS']
        self.SSNList = self.FIBER_SETUPs[CRCM]['SSN']

        self.sciTbl = self.FIBER_SETUPs[CRCM][numpy.where(self.FIBER_SETUPs[CRCM]['TYPE'] != 'S')]
        self.skyTbl = self.FIBER_SETUPs[CRCM][numpy.where(self.FIBER_SETUPs[CRCM]['TYPE'] == 'S')]
        #print(self.tblLists[CRCM]['OzPoz_table'].keys())
        #print(self.tblLists[CRCM]['FLAMES FIBRE Table'].keys())
        #self.sciTbl = self.tblLists[CRCM]['OzPoz_table'][numpy.where(self.tblLists[CRCM]['OzPoz_table']['TYPE'] != 'S')]
        #print(self.sciTbl)
        #self.skyTbl = self.tblLists[CRCM]['OzPoz_table'][numpy.where(self.tblLists[CRCM]['OzPoz_table']['TYPE'] == 'S')]
        #print(self.skyTbl)
        if self.isARG[CRCM] :
            self.sciObjectList = self.sciTbl['Retractor']
        else :
            self.sciObjectList = self.sciTbl['OBJECT']
        self.skyList = self.skyTbl['OBJECT']
        self.objectSkyByName=numpy.append(self.sciObjectList[numpy.argsort(self.sciObjectList)],self.skyList[numpy.argsort(self.skyList)])

    def init_spectrumdisplay_overplot(self, subplot, CRCM='NONE', color='b') :
        op=self.spectrumdisplay.overplot(subplot, self.wave[CRCM], self.flux[CRCM], errorflux=self.err[CRCM], color = color)
        op[0].set_data([],[])
        op[1].set_verts([])
        return op
        
    def init_spec_subplot(self, subplot, CRCM='NONE') :

        wlUnit=""
        try :
            wlUnit=" ["+self.extractedscience[CRCM].all_hdu[0].header['CUNIT2']+"]"
        except :
            pass
        subplot.cla()
        self.spectrumdisplay.setLabels('Wavelength'+wlUnit, 'Total Flux ['+self.bunit[CRCM]+']')
        self.MasterDisplay=self.spectrumdisplay.display(
            subplot = subplot,
            title = "",
            tooltip = "",
            wave = self.wave[CRCM],
            flux = self.flux[CRCM],
            errorflux = self.err[CRCM],
            autolimits = True,
            color = 'b',
        )
        #self.MasterDisplay[1].set_verts(self.errVerts(self.wave[CRCM],self.flux[CRCM],self.err[CRCM]))
        self.MasterDisplay[1].set_verts([])
        self.MasterDisplay[0].set_data([],[])

        self.lineSignal=self.spectrumdisplay.overplot(subplot, [], [], errorflux=[], color = 'b')
        self.lineSignalMinusSky=self.spectrumdisplay.overplot(subplot, [], [], errorflux=[], color = 'g')
        self.lineSky=self.spectrumdisplay.overplot(subplot, [], [], errorflux=[], color = 'r')
        self.AS_lineSignal=self.spectrumdisplay.overplot(subplot, [], [], errorflux=[], color = 'b')
        self.AS_lineSignalMinusSky=self.spectrumdisplay.overplot(subplot, [], [], errorflux=[], color = 'g')
        self.AS_lineSky=self.spectrumdisplay.overplot(subplot, [], [], errorflux=[], color = 'r')
        self.PC_lineSignal=self.spectrumdisplay.overplot(subplot, [], [], errorflux=[], color = 'b')
        self.PC_lineSignalMinusSky=self.spectrumdisplay.overplot(subplot, [], [], errorflux=[], color = 'g')
        self.PC_lineSky=self.spectrumdisplay.overplot(subplot, [], [], errorflux=[], color = 'r')
        self.MED_lineSignal=self.spectrumdisplay.overplot(subplot, [], [], errorflux=[], color = 'b')
        self.MED_lineSignalMinusSky=self.spectrumdisplay.overplot(subplot, [], [], errorflux=[], color = 'g')
        self.MED_lineSky=self.spectrumdisplay.overplot(subplot, [], [], errorflux=[], color = 'r')
        
    def errVerts(self, wave, flux, err):
        vx=numpy.concatenate([wave,numpy.flipud(wave)])
        vy=numpy.concatenate([flux-err,numpy.flipud(flux+err)])
        return [list(zip(vx,vy))]
        

    def setCutsSubPlot(self, subplot, CRCM='NONE') :

        #self.MasterDisplay[0].set_data(self.wave[CRCM], self.flux[CRCM])
        
        #self.MasterDisplay[0].set_data([],[])
        pass

    def selectBrightest(self, CRCM='NONE'):
        if self.nobj[CRCM] == 1:
            self.fib_id = 1
        median = 0
        for obj in range(self.nobj[CRCM]) :
            new_median = numpy.median(self.extractedscience[CRCM].image[:,obj])
            if new_median > median :
                median = new_median
                self.fib_id = obj + 1

    def setSNRarr(self) :
        self.SNRarr={}
        for CRCM in list(self.extractedscience.keys()) :
            self.SNRarr[CRCM]=numpy.zeros(self.nobj[CRCM])
            if self.nobj[CRCM] == 1:
                self.fib_id = 1
            for obj in range(self.nobj[CRCM]) :
                self.SNRarr[CRCM][obj] = giraf_der_snr.DER_SNR(self.extractedscience[CRCM].image[:,obj])

    def selectHighestSNR(self, CRCM='NONE'):
        self.objectSkyBySNR=numpy.append(self.sciObjectList[numpy.flipud(numpy.argsort(self.SNRarr[CRCM][numpy.where(self.FIBER_SETUPs[CRCM]['TYPE'][:] != "S")]))],
                                         self.skyList[numpy.flipud(numpy.argsort(self.SNRarr[CRCM][numpy.where(self.FIBER_SETUPs[CRCM]['TYPE'][:] == "S")]))])
        self.objectSkyByName_ind=numpy.append(numpy.arange(self.nobj[CRCM])[numpy.where(self.FIBER_SETUPs[CRCM]['TYPE'][:] != "S")][numpy.argsort(self.sciObjectList)],
                                              numpy.arange(self.nobj[CRCM])[numpy.where(self.FIBER_SETUPs[CRCM]['TYPE'][:] == "S")][numpy.argsort(self.skyList)])
        self.objectSkyBySNR_ind=numpy.append(numpy.arange(self.nobj[CRCM])[numpy.where(self.FIBER_SETUPs[CRCM]['TYPE'][:] != "S")][numpy.flipud(numpy.argsort(self.SNRarr[CRCM][numpy.where(self.FIBER_SETUPs[CRCM]['TYPE'][:] != "S")]))],
                                             numpy.arange(self.nobj[CRCM])[numpy.where(self.FIBER_SETUPs[CRCM]['TYPE'][:] == "S")][numpy.flipud(numpy.argsort(self.SNRarr[CRCM][numpy.where(self.FIBER_SETUPs[CRCM]['TYPE'][:] == "S")]))])

        # Select Highest SNR
        if(self.fib_id == -1) :
            self.fib_id = numpy.argsort(self.SNRarr[CRCM])[-1]+1

        self.skyBySNR_ind=numpy.arange(self.nobj[CRCM])[numpy.where(self.FIBER_SETUPs[CRCM]['TYPE'][:] == "S")][numpy.flipud(numpy.argsort(self.SNRarr[CRCM][numpy.where(self.FIBER_SETUPs[CRCM]['TYPE'][:] == "S")]))]
        self.skyListSNR=self.SNRarr[CRCM][self.skyBySNR_ind]

    def setFluxSelected(self) :
        self.flux = {}
        self.err = {}
        self.OBJECT = {}
        self.FPS = {}
        self.SSN = {}
        self.OBJECT_SNR = {}
        for CRCM in list(self.extractedscience.keys()) :
            self.flux[CRCM] = self.extractedscience[CRCM].image[:,self.fib_id-1]
            self.err[CRCM] = self.extractederrors[CRCM].image[:,self.fib_id-1]
            self.OBJECT[CRCM] = self.FIBER_SETUPs[CRCM]['OBJECT'][self.fib_id-1]
            self.FPS[CRCM] = self.FIBER_SETUPs[CRCM]['FPS'][self.fib_id-1]
            self.SSN[CRCM] = self.FIBER_SETUPs[CRCM]['SSN'][self.fib_id-1]
            if self.isARG[CRCM] and not self.FIBER_SETUPs[CRCM]['TYPE'][self.fib_id-1] == 'S' :
                self.OBJECT[CRCM] = self.extractedscience[CRCM].all_hdu[0].header['HIERARCH ESO OBS TARG NAME']
            self.OBJECT_SNR[CRCM] = self.SNRarr[CRCM][self.fib_id-1]

    def selectObject(self, fib_id):
        self.fib_id = fib_id
        self.setFluxSelected()

    def plot(self, subplot, title, tooltip, CRCM='NONE'):

        # Plot the data...
        ltitle='Rebinned spectrum of %s, S/N = %5.2f [FPS/SSN=%d/%d, col=%d]' %(self.OBJECT[CRCM].strip(), self.OBJECT_SNR[CRCM], self.FPS[CRCM], self.SSN[CRCM], self.fib_id)
        # --------------------------------------------------------------------------------
        ## CRCM = NONE
        if self.flux is None :
            return
        if 'NONE' in list(self.flux.keys()) :
            if self.NoCRC_displaySignal :
                self.lineSignal[0].set_data(self.wave['NONE'], self.flux['NONE'])
                if self.NoCRC_displayErr :
                    self.lineSignal[1].set_verts(self.errVerts(self.wave['NONE'], self.flux['NONE'], self.err['NONE']))
                else :
                    self.lineSignal[1].set_verts([])
            else :
                self.lineSignal[0].set_data([], [])
                self.lineSignal[1].set_verts([])
            
            # Overplot the sky?
            if not self.skyApplyTo is None and not self.sky['NONE'] is None and self.NoCRC_displaySky :
                self.lineSky[0].set_data(self.wave['NONE'], self.sky['NONE'][0][:,self.fib_id-1])
                if self.NoCRC_displayErr :
                    self.lineSky[1].set_verts(self.errVerts(self.wave['NONE'], self.sky['NONE'][0][:,self.fib_id-1], self.sky['NONE'][1][:,self.fib_id-1]))
                else :
                    self.lineSky[1].set_verts([])
            else :
                self.lineSky[0].set_data([],[])
                self.lineSky[1].set_verts([])
            # Overplot the spectrum-sky?
            if not self.skyApplyTo is None and not self.sky['NONE'] is None and self.NoCRC_displaySignalMinusSky :
                self.lineSignalMinusSky[0].set_data(self.wave['NONE'], self.flux['NONE']-self.sky['NONE'][0][:,self.fib_id-1])
                if self.NoCRC_displayErr :
                    self.lineSignalMinusSky[1].set_verts(self.errVerts(self.wave['NONE'], self.flux['NONE']-self.sky['NONE'][0][:,self.fib_id-1], self.err['NONE']+self.sky['NONE'][1][:,self.fib_id-1]))
                else :
                    self.lineSignalMinusSky[1].set_verts([])
            else :
                self.lineSignalMinusSky[0].set_data([],[])
                self.lineSignalMinusSky[1].set_verts([])
        # --------------------------------------------------------------------------------
        ## CRCM = astroSCRAPPY
        if 'ASTROSCRAPPY' in list(self.flux.keys()) :
            if self.AS_displaySignal :
                self.AS_lineSignal[0].set_data(self.wave['ASTROSCRAPPY'], self.flux['ASTROSCRAPPY'])
                if self.AS_displayErr :
                    self.AS_lineSignal[1].set_verts(self.errVerts(self.wave['ASTROSCRAPPY'], self.flux['ASTROSCRAPPY'], self.err['ASTROSCRAPPY']))
                else :
                    self.AS_lineSignal[1].set_verts([])
            else :
                self.AS_lineSignal[0].set_data([], [])
                self.AS_lineSignal[1].set_verts([])
            
            # Overplot the sky?
            if not self.skyApplyTo is None and not self.sky['ASTROSCRAPPY'] is None and self.AS_displaySky :
                self.AS_lineSky[0].set_data(self.wave['ASTROSCRAPPY'], self.sky['ASTROSCRAPPY'][0][:,self.fib_id-1])
                if self.AS_displayErr :
                    self.AS_lineSky[1].set_verts(self.errVerts(self.wave['ASTROSCRAPPY'], self.sky['ASTROSCRAPPY'][0][:,self.fib_id-1], self.sky['ASTROSCRAPPY'][1][:,self.fib_id-1]))
                else :
                    self.AS_lineSky[1].set_verts([])
            else :
                self.AS_lineSky[0].set_data([],[])
                self.AS_lineSky[1].set_verts([])
            # Overplot the spectrum-sky?
            if not self.skyApplyTo is None and not self.sky['ASTROSCRAPPY'] is None and self.AS_displaySignalMinusSky :
                self.AS_lineSignalMinusSky[0].set_data(self.wave['ASTROSCRAPPY'], self.flux['ASTROSCRAPPY']-self.sky['ASTROSCRAPPY'][0][:,self.fib_id-1])
                if self.AS_displayErr :
                    self.AS_lineSignalMinusSky[1].set_verts(self.errVerts(self.wave['ASTROSCRAPPY'], self.flux['ASTROSCRAPPY']-self.sky['ASTROSCRAPPY'][0][:,self.fib_id-1], self.err['ASTROSCRAPPY']+self.sky['ASTROSCRAPPY'][1][:,self.fib_id-1]))
                else :
                    self.AS_lineSignalMinusSky[1].set_verts([])
            else :
                self.AS_lineSignalMinusSky[0].set_data([],[])
                self.AS_lineSignalMinusSky[1].set_verts([])
        # --------------------------------------------------------------------------------
        ## CRCM = PyCosmic
        if 'PYCOSMIC' in list(self.flux.keys()) :
            if self.PC_displaySignal :
                self.PC_lineSignal[0].set_data(self.wave['PYCOSMIC'], self.flux['PYCOSMIC'])
                if self.PC_displayErr :
                    self.PC_lineSignal[1].set_verts(self.errVerts(self.wave['PYCOSMIC'], self.flux['PYCOSMIC'], self.err['PYCOSMIC']))
                else :
                    self.PC_lineSignal[1].set_verts([])
            else :
                self.PC_lineSignal[0].set_data([], [])
                self.PC_lineSignal[1].set_verts([])
            
            # Overplot the sky?
            if not self.skyApplyTo is None and not self.sky['PYCOSMIC'] is None and self.PC_displaySky :
                self.PC_lineSky[0].set_data(self.wave['PYCOSMIC'], self.sky['PYCOSMIC'][0][:,self.fib_id-1])
                if self.PC_displayErr :
                    self.PC_lineSky[1].set_verts(self.errVerts(self.wave['PYCOSMIC'], self.sky['PYCOSMIC'][0][:,self.fib_id-1], self.sky['PYCOSMIC'][1][:,self.fib_id-1]))
                else :
                    self.PC_lineSky[1].set_verts([])
            else :
                self.PC_lineSky[0].set_data([],[])
                self.PC_lineSky[1].set_verts([])
            # Overplot the spectrum-sky?
            if not self.skyApplyTo is None and not self.sky['PYCOSMIC'] is None and self.PC_displaySignalMinusSky :
                self.PC_lineSignalMinusSky[0].set_data(self.wave['PYCOSMIC'], self.flux['PYCOSMIC']-self.sky['PYCOSMIC'][0][:,self.fib_id-1])
                if self.PC_displayErr :
                    self.PC_lineSignalMinusSky[1].set_verts(self.errVerts(self.wave['PYCOSMIC'], self.flux['PYCOSMIC']-self.sky['PYCOSMIC'][0][:,self.fib_id-1], self.err['PYCOSMIC']+self.sky['PYCOSMIC'][1][:,self.fib_id-1]))
                else :
                    self.PC_lineSignalMinusSky[1].set_verts([])
            else :
                self.PC_lineSignalMinusSky[0].set_data([],[])
                self.PC_lineSignalMinusSky[1].set_verts([])
        ## CRCM = Median
        if 'MEDIAN' in list(self.flux.keys()) :
            if self.MED_displaySignal :
                self.MED_lineSignal[0].set_data(self.wave['MEDIAN'], self.flux['MEDIAN'])
                if self.MED_displayErr :
                    self.MED_lineSignal[1].set_verts(self.errVerts(self.wave['MEDIAN'], self.flux['MEDIAN'], self.err['MEDIAN']))
                else :
                    self.MED_lineSignal[1].set_verts([])
            else :
                self.MED_lineSignal[0].set_data([], [])
                self.MED_lineSignal[1].set_verts([])
            
            # Overplot the sky?
            if not self.skyApplyTo is None and not self.sky['MEDIAN'] is None and self.MED_displaySky :
                self.MED_lineSky[0].set_data(self.wave['MEDIAN'], self.sky['MEDIAN'][0][:,self.fib_id-1])
                if self.MED_displayErr :
                    self.MED_lineSky[1].set_verts(self.errVerts(self.wave['MEDIAN'], self.sky['MEDIAN'][0][:,self.fib_id-1], self.sky['MEDIAN'][1][:,self.fib_id-1]))
                else :
                    self.MED_lineSky[1].set_verts([])
            else :
                self.MED_lineSky[0].set_data([],[])
                self.MED_lineSky[1].set_verts([])
            # Overplot the spectrum-sky?
            if not self.skyApplyTo is None and not self.sky['MEDIAN'] is None and self.MED_displaySignalMinusSky :
                self.MED_lineSignalMinusSky[0].set_data(self.wave['MEDIAN'], self.flux['MEDIAN']-self.sky['MEDIAN'][0][:,self.fib_id-1])
                if self.MED_displayErr :
                    self.MED_lineSignalMinusSky[1].set_verts(self.errVerts(self.wave['MEDIAN'], self.flux['MEDIAN']-self.sky['MEDIAN'][0][:,self.fib_id-1], self.err['MEDIAN']+self.sky['MEDIAN'][1][:,self.fib_id-1]))
                else :
                    self.MED_lineSignalMinusSky[1].set_verts([])
            else :
                self.MED_lineSignalMinusSky[0].set_data([],[])
                self.MED_lineSignalMinusSky[1].set_verts([])
        self.MasterDisplay[2].set_text(ltitle)


    def computeSky(self, applyTo, method, statistic, numFibs=None ):
        self.sky={}
        for CRCM in list(self.extractedscience.keys()) :
            self.sky[CRCM] = self.computeSkyCRCM(applyTo, method, statistic, numFibs=numFibs, CRCM=CRCM)
        return self.sky

    def copyCRCkwsToHeader(self, HDU, CRCM="NONE"):
        ## create list of keywords with:
        # grep 'header.append(.*' giraf_astroscrappy_multi.py giraf_wfr_reflex_{median,pycosmic}_multi.py | gsed -e 's/.*header.append((\(.*\),.*/\1/' -e "s/'/:/g" -e 's/:/"/g' -e 's/^/          /' -e 's/,.*//' -e 's/$/,/'| sort -u
        for K in [
          "CMETHOD",
          "CRBPM",
          "CRGAIN",
          "CRGFWHM",
          "CRITER",
          "CRMBIAS",
          "CRRADIUS",
          "CRRBOX",
          "CRRLIM",
          "CRRON",
          "CRSIGLIM",
          "HIERARCH ESO DRS ASCRAP BIAS NAME",
          "HIERARCH ESO DRS ASCRAP BIASSUB",
          "HIERARCH ESO DRS ASCRAP GAIN",
          "HIERARCH ESO DRS ASCRAP RON",
          "HIERARCH ESO DRS ASCRAP",
          "HIERARCH ESO DRS CMETHOD",
          "HIERARCH ESO DRS PYCOSMIC BIAS NAME",
          "HIERARCH ESO DRS PYCOSMIC BIASSUB",
          "HIERARCH ESO DRS PYCOSMIC GAIN",
          "HIERARCH ESO DRS PYCOSMIC RON",
          "HIERARCH ESO DRS PYCOSMIC",
          "NFILES",        
          ] :
              if K in self.extractedscience[CRCM].all_hdu[0].header :
                  HDU.header[K]=self.extractedscience[CRCM].all_hdu[0].header[K]
                  if K == "NFILES" :
                      for i in range(self.extractedscience[CRCM].all_hdu[0].header[K]) :
                          if "IFILES"+str(i) in self.extractedscience[CRCM].all_hdu[0].header :
                              HDU.header["IFILES"+str(i)]=self.extractedscience[CRCM].all_hdu[0].header["IFILES"+str(i)]
                              

    def updateSkyHeader(self, HDU, CRCM="NONE"):
        HDU.header["HIERARCH ESO DRS SKY APPLYTO"]=self.skyApplyTo
        HDU.header["HIERARCH ESO DRS SKY METHOD"]=self.skyMethod
        HDU.header["HIERARCH ESO DRS SKY STATISTIC"]=self.skyStatistic
        HDU.header["HIERARCH ESO DRS SKY NUMFIBS"]=self.skyNumFibs
        HDU.header["HIERARCH ESO DRS SKY REJECT"]=numpy.any(self.skyRejected == False)
        j=0
        for i in range(len(self.FIBER_SETUPs[CRCM][self.skyBySNR_ind])) :
            HDU.header["HIERARCH ESO DRS SKY IDX%03d ID" %(self.skyBySNR_ind[i])]=self.skyList[i]
            HDU.header["HIERARCH ESO DRS SKY IDX%03d SNR" %(self.skyBySNR_ind[i])]=self.skyListSNR[i]
            HDU.header["HIERARCH ESO DRS SKY IDX%03d ISSEL" %(self.skyBySNR_ind[i])]=self.skyFibIsSelected[i]
            if self.skyBySNR_ind[i] in self.skyBySNR_ind[self.skyFibIsSelected] :
                HDU.header["HIERARCH ESO DRS SKY IDX%03d ISREJ" %(self.skyBySNR_ind[i])]=not self.skyRejected[j]
                j+=1
            else :
                HDU.header["HIERARCH ESO DRS SKY IDX%03d ISREJ" %(self.skyBySNR_ind[i])]=True

    def computeSkyCRCM(self, applyTo, method, statistic, numFibs=None, CRCM='NONE' ):
        sky=None
        self.skyApplyTo = applyTo
        self.skyMethod = method
        self.skyStatistic = statistic
        self.skyNumFibs = numFibs

        if applyTo == "None" :
            return sky

        cs_OneSkyValue=False
        if method == "None" :
            return sky
        elif method == "All Sky Fibres" :
            cs_skyfibs = self.skyBySNR_ind[self.skyFibIsSelected][self.skyRejected]
            cs_fibrange = [self.fib_id-1,]
            cs_OneSkyValue=True
        elif method == "Best <N> Fibres by S/N" :
            cs_skyfibs = self.skyBySNR_ind[self.skyFibIsSelected][self.skyRejected][0:numFibs]
            cs_fibrange = [self.fib_id-1,]
            cs_OneSkyValue=True
        elif method == "Nearest <N> Fibres" :
            if numFibs == len(self.skyList) :
                cs_skyfibs = self.skyBySNR_ind[self.skyFibIsSelected][self.skyRejected]
                cs_fibrange = [self.fib_id-1,]
                cs_OneSkyValue=True
            else :
                cs_fibrange = list(range(self.nobj[CRCM]))
                if self.isARG[CRCM] :
                    tc=Table([[self.RADEC.ra.deg,],[self.RADEC.dec.deg,]], names=("RA","DEC"))
                else:
                    tc=self.FIBER_SETUPs[CRCM]
                otList = searchInCat(tc, self.FIBER_SETUPs[CRCM][self.skyBySNR_ind][self.skyFibIsSelected][self.skyRejected], rad=[0.,30.*60.], maxsources=numFibs)
        elif method == "Improves S/N" :
            print(("method = %s Not yet implemented" %(method)))
        else :
            print(("Error, method = %s not supported" %(method)))
            return sky

        if applyTo == "All" :
            # Apply to all science spectra
            pass
        elif applyTo == "One" :
            print(("applyTo = %s Not yet implemented" %(applyTo)))
            # Apply to only the selected science spectra
        else :
            print(("Error, appyTo = %s not supported" %(appyTo)))
            return sky

        if not self.sky is None and CRCM in list(self.sky.keys()) :
            sky=self.sky[CRCM]
        else :
            sky=numpy.array([numpy.zeros(numpy.shape(self.extractedscience[CRCM].image)),numpy.zeros(numpy.shape(self.extractederrors[CRCM].image)),])

        for fib in cs_fibrange :
            if method == "Nearest <N> Fibres" :
                if numFibs != len(self.skyList) :
                    if fib+1 in self.skyTbl["INDEX"] :
                        # For a sky fibre the closest sky fib will always be itself, so exclude that one...
                        cs_skyfibs=otList["skys"][numpy.where(otList["skys"]["_idx"] == fib)]["INDEX"][1:numFibs+1]-1
                    else :
                        cs_skyfibs=otList["skys"][numpy.where(otList["skys"]["_idx"] == fib)]["INDEX"][0:numFibs]-1
            if statistic == "None" :
                sky=None
                return sky
            elif statistic.lower() == "median" :
                cs_sky=numpy.array([
                  numpy.median(self.extractedscience[CRCM].image[:,cs_skyfibs],1),
                  numpy.mean(self.extractederrors[CRCM].image[:,cs_skyfibs],1),
                ])
            elif statistic.lower() == "mean" :
                cs_sky=numpy.array([
                  numpy.mean(self.extractedscience[CRCM].image[:,cs_skyfibs],1),
                  numpy.mean(self.extractederrors[CRCM].image[:,cs_skyfibs],1),
                ])
            elif statistic.lower() == "mode" :
                print(("statistic = %s Not yet implemented" %(statistic)))
                ## If ever to implement, probably best to use scipy.statsi.mode(),see http://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.mode.html#scipy.stats.mode
                ## and https://stackoverflow.com/questions/16330831/most-efficient-way-to-find-mode-in-numpy-array
            else :
                print(("Error, statistic = %s not supported" %(statistic)))
                sky=None
                return sky

            if applyTo == "All" and cs_OneSkyValue :
                # Apply to all science spectra
                sky=numpy.array([
                  numpy.transpose(cs_sky[0]*numpy.transpose(numpy.ones(numpy.shape(self.extractedscience[CRCM].image)))),
                  numpy.transpose(cs_sky[1]*numpy.transpose(numpy.ones(numpy.shape(self.extractederrors[CRCM].image)))),
                ])
            elif applyTo == "One" :
                #print("applyTo = %s Not yet implemented" %(applyTo))
                # Apply to only the selected science spectra
                sky[0][self.fib_id-1]=cs_sky[0]
                sky[1][self.fib_id-1]=cs_sky[1]
            elif not cs_OneSkyValue :
                sky[0][:,fib-1]=cs_sky[0]
                sky[1][:,fib-1]=cs_sky[1]
#        else :
#            print("Error, appyTo = %s not supported" %(applyTo))
#            sky=None
#            return sky

        #print(numpy.mean(sky[0][:,self.fib_id-1]))
        #print(numpy.mean(sky[1][:,self.fib_id-1]))
        return sky
################################################################################
def searchInCat(tc, catTable, rad, _maxMag=None, _magCol=None, _frame="icrs", maxsources=10000 ):

    #catroot, catroot, defMagCol, dict_magCol, colNames = vizierVars(catName)

    if _maxMag != None :
        maxMag=numpy.array(_maxMag.split(','), dtype='|S30').astype(numpy.float)
    else :
        maxMag=numpy.array([99.,])
    magCol = _magCol

    if magCol == None :
        magCol = 'Mag'

    maxSources=numpy.array(maxsources).astype(numpy.float)

    # How do I get the frame of the catalog...???
    catalog = SkyCoord(catTable["RA"],catTable["DEC"])

    if numpy.isscalar(tc["RA"]) :
        c=SkyCoord([tc["RA"],],[tc["DEC"],], unit=(u.deg, u.deg), frame=_frame)
    else :
        c=SkyCoord(tc["RA"],tc["DEC"], unit=(u.deg, u.deg), frame=_frame)

    if len(tc) > 0 :
        idxc, idxcatalog, d2d, d3d = catalog.search_around_sky(c, rad[1]*u.arcsec)
        idx=numpy.where(d2d > rad[0]*u.arcsec)
        sidxc=idxc[idx]
        sidxcatalog=idxcatalog[idx]
        st=catTable[sidxcatalog]
        sd2d=d2d[idx]
        if _maxMag != None :
            idx=numpy.where(st[dict_magCol[magCol]] < maxMag[0])
            st=st[idx]
            sidxc=sidxc[idx]
            sidxcatalog=sidxcatalog[idx]
            sd2d=sd2d[idx]
            if len(maxMag) > 1 :
                idx=numpy.where(st[dict_magCol[magCol]] > maxMag[1])
                st=st[idx]
                sidxc=sidxc[idx]
                sidxcatalog=sidxcatalog[idx]
                sd2d=sd2d[idx]
        #st.sort("_r")
        # Matched objects in input Coordinates
        mc=c[numpy.unique(sidxc)]
        # UnMatched objects in input Coordinates
        umcidx = [x for x in range(len(c)) if x not in sidxc]
        umc=c[umcidx]

        st['_idx']=sidxc
        st['_r']=sd2d.deg
        st['_r'].unit=u.deg
        tmc=Table()
        tmc['RA']=mc.ra.deg
        tmc['DEC']=mc.dec.deg
        tmc['_idx']=numpy.unique(sidxc)
        tumc=Table()
        tumc['RA']=umc.ra.deg
        tumc['DEC']=umc.dec.deg
        tumc['_idx']=umcidx

        #return TableList([("skys",st),('matched',tmc),('unmatched',tumc)]),rad,maxMag,magCol,maxsources
        #return TableList([('skys',st),('matched',tmc),('unmatched',tumc)])
        return {'skys': st, 'matched': tmc, 'unmatched': tumc}

################################################################################



# ========================================================================================================
#
# John Pritchard
#
# Developed from esotk_fringe_obj_mask.py by Mark Neeser
#
# A python interactive window for use in the Austrian in-kind
# fringe-correction Reflex workflow.  This interactive window
# shows the results of the fringe-correction pipeline, allowing
# the User to compare the input image with the fringe-corrected
# image, and the fringe map.   From this interactive Actor the
# User can modify pipeline parameters and re-initiate the
# processing.
#
#
#
# ========================================================================================================

import os
import re
import sys
sys.path.append(os.path.dirname(__file__))
from esotk_giraf_gui import BaseDataPlotterManager, runInteractiveApp

try:
    # Import non-standard but needed modules. These must be as subset of what
    # the esotk_fringe_gui module tries to load.
    import reflex
    import pipeline_product
    import pipeline_display
    import matplotlib
    from giraf_plot_common import *

    class DataPlotterManager(BaseDataPlotterManager):

        # This function will read all the columns, images and whatever is needed
        # from the products. The variables , self.plot_x, self.plot_y, etc...
        # are used later in function plotProductsGraphics().
        def readFitsData(self, fitsFiles):
            
            # If fitsFiles doesn't have FF_PSFCENTROID, recipe must have failed, so search for tmp products...
            #Read all the products
            frames = dict()
            for frame in fitsFiles :
                if frame == '' :
                    continue
                category = frame.category
                frames[category] = frame
            self.have_FF_PSFCENTROID=True
            if not "FF_PSFCENTROID" in frames :
                self.have_FF_PSFCENTROID=False
                import glob
                for F in glob.glob('../../gimasterflat_1/latest/*.fits') :
                    HDUlist=pyfits.open(F)
                    try :
                        fitsFiles.append(reflex.FitsFile(os.path.abspath(F), HDUlist[0].header['HIERARCH ESO PRO CATG'], None, None))
                    except :
                        pass
                    HDUlist.close()

            # Control variable to check if the interesting files were found in
            # the input.
            self.flat_frames = []
            self.science_frames = []
            self.raw_frames = []
            self.frameLabels = {}

            #Initialise the objects to read/display
            self.masterflat     = None
            self.ff_loccentroid = None
            self.ff_psfcentroid = None
            self.ff_raw = None
            self.objid = 1

            # Read all the products
            ff = {}
            for frame in fitsFiles:
                if frame == '':
                    continue
                category = frame.category
                if category == 'MASTER_FIBER_FLAT':
                    self.frameLabels[frame.name]=frame.category+': '+frame.name
                    ff[frame.category]=frame
                    self.masterflat   = frame
                elif category == 'FF_LOCCENTROID':
                    self.frameLabels[frame.name]=frame.category+': '+frame.name
                    ff[frame.category]=frame
                    self.ff_loccentroid = frame
                elif category == 'FF_PSFCENTROID':
                    self.frameLabels[frame.name]=frame.category+': '+frame.name
                    ff[frame.category]=frame
                    self.ff_psfcentroid = frame
                elif category == 'FIBER_FLAT':
                    self.raw_frames.append(frame)
            if self.ff_psfcentroid is not None :
                self.science_frames.append(ff['FF_PSFCENTROID'])
            if self.ff_loccentroid is not None :
                self.science_frames.append(ff['FF_LOCCENTROID'])
            elif self.masterflat is not None :
                self.science_frames.append(ff['MASTER_FIBER_FLAT'])

            if self.ff_psfcentroid is not None :
                self.masterflatPlot   = PlotableFlatWithPSFcentroid(self.masterflat,self.ff_psfcentroid)
            elif self.ff_loccentroid is not None :
                self.masterflatPlot   = PlotableFlatWithLOCcentroid(self.masterflat,self.ff_loccentroid)
            elif self.masterflat is not None :
                self.masterflatPlot   = PlotableFlat(self.masterflat)

            self.science_found = len(self.science_frames) > 0
            self.raw_found = len(self.raw_frames) > 0

            self.DPR_CATG = "unknown"
            self.DPR_TYPE = "unknown"
            self.DPR_TECH = "unknown"
            if self.raw_found:
                HDUlist=pyfits.open(self.raw_frames[0].name)
                self.DPR_CATG = HDUlist[0].header['HIERARCH ESO DPR CATG']
                self.DPR_TYPE = HDUlist[0].header['HIERARCH ESO DPR TYPE']
                self.DPR_TECH = HDUlist[0].header['HIERARCH ESO DPR TECH']
                HDUlist.close()

            if self.science_found:
                self.esorex_log_file_name  = re.sub(".*/giraffe/" ,"../../", os.path.dirname(self.science_frames[0].name))+"/log_dir/esorex.log"

        def loadProducts(self, index):
            """
            Loading the plots...
            """
            super(DataPlotterManager, self).loadProducts(index)
            if self.masterflat is not None :
                if self.science_frames[index].category == 'FF_PSFCENTROID' :
                    self.masterflatPlot   = PlotableFlatWithPSFcentroid(self.masterflat,self.ff_psfcentroid)
                elif self.science_frames[index].category == 'FF_LOCCENTROID' :
                    self.masterflatPlot   = PlotableFlatWithLOCcentroid(self.masterflat,self.ff_loccentroid)
                elif self.science_frames[index].category == 'MASTER_FIBER_FLAT' :
                    self.masterflatPlot   = PlotableFlat(self.science_frames[index])
                
        # This function creates all the subplots. It is responsible for the
        # plotting layouts.
        # There can different layouts, depending on the availability of data
        # Note that subplot(I,J,K) means the Kth plot in a IxJ grid
        # Note also that the last one is actually a box with text, no graphs.
        def addSubplots(self, figure):
            self.list_subplot_image     = []
            if self.masterflat is not None :
                if self.ff_loccentroid is not None :
                    gs = matplotlib.gridspec.GridSpec(100,100)
                    self.subplot_masterflat    = figure.add_subplot(gs[0:43,0:100])
                    self.subplot_psfcentroid   = figure.add_subplot(gs[54:92,0:100])
                    self.subplot_txtinfo       = figure.add_subplot(gs[95:100,0:100])
                    #self.subplot_txtinfo       = figure.widgets.Widget(gs[95:100,0:100])
                else :
                    gs = matplotlib.gridspec.GridSpec(100,100)
                    self.subplot_masterflat    = figure.add_subplot(gs[0:72,0:100])
                    self.subplot_psfcentroid   = figure.add_subplot(gs[80:92,0:100])
                    self.subplot_txtinfo       = figure.add_subplot(gs[95:100,0:100])

                self.setupFigure(figure)
            else :
                self.subtext_nodata      = figure.add_subplot(1,1,1)

        # This is the function that makes the plots.
        # Add new plots or delete them using the given scheme.
        # The data has been already stored in self.plot_x, self.plot_xdif, etc ...
        # It is mandatory to add a tooltip variable to each subplot.
        # One might be tempted to merge addSubplots() and plotProductsGraphics().
        # There is a reason not to do it: addSubplots() is called only once at
        # startup, while plotProductsGraphics() is called always there is a
        # resize.
        def plotProductsGraphics(self):

            #if self.science_found == True and self.err_found == True:
                
            if self.masterflat is not None :

                self.plotMasterFlatPlot()
                if self.ff_loccentroid is not None :
                    if self.ff_psfcentroid is None :
                        self.showNoPSF()
                    elif not self.have_FF_PSFCENTROID :
                        self.showHavePSFbutFAILED()

                else:
                    self.showNoLOC()

                #Additional text info
                self.showTextInfo()

            else :
                #Data not found info
                self.showNoData()

        def plotMasterFlatPlot(self) :
            if type(self.masterflatPlot) is PlotableFlat :
                title_masterflat   = 'Master Flat'
                tooltip_masterflat = """Master Flat.
The master flat generated by the gimasterflat recipe, apparently processing failed, probably the recipe was
unable to find the expected number of fibres."""
                self.masterflatPlot.plot(self.subplot_masterflat, title_masterflat,
                                 tooltip_masterflat)

            elif type(self.masterflatPlot) is PlotableFlatWithLOCcentroid :
                title_masterflat   = 'Master Flat with LOCCENTROID'
                tooltip_masterflat = """Master Flat.
The centroids of the fibre location overlaid on the master flat."""
                self.masterflatPlot.plot(self.subplot_masterflat, title_masterflat,
                                 tooltip_masterflat)
                title_masterflat   = 'Master Flat cross cut with LOCCENTROID'
                tooltip_masterflat = """Master Flat.
The centroids of the fibre location overlaid on a cross section of the master flat."""
                self.masterflatPlot.plotXsection(self.subplot_psfcentroid, title_masterflat,
                                 tooltip_masterflat)

            elif type(self.masterflatPlot) is PlotableFlatWithPSFcentroid :
                title_masterflat   = 'Master Flat with PSFCENTROID'
                tooltip_masterflat = """Master Flat.
The centroids of the PSF fitting overlaid on the master flat."""
                self.masterflatPlot.plot(self.subplot_masterflat, title_masterflat,
                                 tooltip_masterflat)
                title_masterflat   = 'Master Flat cross cut with PSFCENTROID'
                tooltip_masterflat = """Master Flat.
The centroids of the PSF fitting overlaid on a cross section of the master flat."""
                self.masterflatPlot.plotXsection(self.subplot_psfcentroid, title_masterflat,
                                 tooltip_masterflat)

        def showNoLOC(self) :
          #Data not found info
          self.subplot_psfcentroid.set_axis_off()
          self.text_nodata = 'Recipe FAILED, could not find product FF_LOCCENTROID.'
          self.subplot_psfcentroid.text(0.5, 0.5, self.text_nodata, color='red', fontsize=18,
                                   ha='center', va='center', alpha=1.0)
          self.subplot_psfcentroid.tooltip="""The gimasterflat recipe has apparently failed to even localaize the fibres.
Probably it failed to locate the expected number of fibres, try adjusting fiber-splist on the 'Fibre selection' tab, and/or the
parameters on the 'Location' tab."""

        def showNoPSF(self) :
          #Data not found info
          self.subplot_psfcentroid.set_axis_off()
          self.text_nodata = 'Recipe FAILED, found FF_LOCCENTROID product,\nbut could not find product FF_PSFCENTROID.'
          self.subplot_psfcentroid.text(0., 0., self.text_nodata, color='red', fontsize=16,
                                   ha='left', va='top', alpha=1.0)
          self.subplot_psfcentroid.tooltip="""The gimasterflat recipe has apparently failed to build the PSF models of fibres.
Possibly the localization found the right number of fibres, but failed to trace them correctly. Try adjusting the
parameters on the 'Location' and 'PSF' tab."""


        def showHavePSFbutFAILED(self) :
          #data found in working dir, but recipe still failed...
          self.subplot_psfcentroid.set_axis_off()
          self.text_nodata = 'Found FF_LOCCENTROID & FF_PSFCENTROID but recipe still FAILED\nPlease check the log file.'
          self.subplot_psfcentroid.text(0., 0., self.text_nodata, color='red', fontsize=16,
                                   ha='left', va='top', alpha=1.0)
          self.subplot_psfcentroid.tooltip="""The gimasterflat recipe has apparently failed to complete successfully.
Possibly the localization found the right number of fibres, but failed to trace them correctly. Try adjusting the
parameters on the 'Location' and 'PSF' tab."""


        def showTextInfo(self) :
            self.subplot_txtinfo.set_axis_off()
            try:
                mode = self.masterflatPlot.flat.readKeyword('HIERARCH ESO INS SLITS ID')
            except:
                mode = 'unknown'
            try:
                filter_name = self.masterflatPlot.flat.readKeyword('HIERARCH ESO INS FILT NAME')
            except:
                filter_name = 'unknown'
            try:
                grat_wlen = self.masterflatPlot.flat.readKeyword('HIERARCH ESO INS GRAT WLEN')
            except:
                grat_wlen = 'unknown'
            try:
                raw1_name = self.masterflatPlot.flat.readKeyword('HIERARCH ESO PRO REC1 RAW1 NAME')
            except:
                raw1_name = 'unknown'

            self.subplot_txtinfo.text(0.1, 0., 'DPid RAW1: %s\nDPR: %s :: %s :: %s\nMode/Setting/WLen: %s/%s/%s'
                                   %(raw1_name,
                                     self.DPR_CATG,self.DPR_TYPE,self.DPR_TECH,
                                     mode,filter_name,str(grat_wlen)
                                    ),
                                   ha='left', va='top', weight='bold')

            try:
                slit = self.sci_extracted.extractedscience.readKeyword('HIERARCH ESO INS SLIT NAME')
                self.subplot_txtinfo.text(0.5, 0., 'Slit name (LSS): '+slit, 
                                   ha='left', va='center', weight='bold')
            except:
                pass
            try:
                checksum = self.sci_extracted.extractedscience.readKeyword('HIERARCH ESO INS MOS CHECKSUM')
                self.subplot_txtinfo.text(0.5, 0., 'Slits checksum (MOS): '+str(checksum), 
                                   ha='left', va='center', weight='bold')
            except:
                pass
            try:
                mask_id = self.sci_extracted.extractedscience.readKeyword('HIERARCH ESO INS MASK ID')
                self.subplot_txtinfo.text(0.5, 0., 'Mask id (MXU): '+mask_id, 
                                   ha='left', va='center', weight='bold')
            except:
                pass
            try:
                target_name = self.sci_extracted.extractedscience.readKeyword('HIERARCH ESO OBS TARG NAME')
                self.subplot_txtinfo.text(0.1, 0.6, 'Target name: '+target_name, 
                                   ha='left', va='center', weight='bold')
            except:
                pass

        def showNoData(self) :
            #Data not found info
            self.subtext_nodata.set_axis_off()
            self.text_nodata = 'Master Flat data not found in the products\n(PRO.CATG=MASTER_FIBER_FLAT)'
            self.subtext_nodata.text(0.1, 0.6, self.text_nodata, color='#11557c', fontsize=18,
                                   ha='left', va='center', alpha=1.0)
            self.subtext_nodata.tooltip='Master Flat data not found in the products'


        def plotWidgets(self) :
            widgets = list()
            if False and self.masterflat is not None :
                # Clickable subplot
                self.clickablesmapped = InteractiveClickableSubplot(
                       self.subplot_masterflat, self.setExtractedObject)
                widgets.append(self.clickablesmapped)
            return widgets

        def setExtractedObject(self, point) :
            obj = self.masterflat.getObjectInPosition(point.ydata)
            if obj != -1 :
                self.sci_extracted.selectObject(obj)
            self.plotSciExtracted()

    
        # This function specifies which are the parameters that should be presented
        # in the window to be edited.
        # Note that the parameter has to be also in the in_sop port (otherwise it
        # won't appear in the window)
        # The descriptions are used to show a tooltip. They should match one to one
        # with the parameter list
        # Note also that parameters have to be prefixed by the 'recipe name:'
        def setInteractiveParameters(self):
            paramList = list()
            paramList.append(reflex.RecipeParameter('gimasterflat','fiber-splist',group='Geom',description='Index list of spectra to use for localization (e.g. 2,10,30-40,55)'))
            paramList.append(reflex.RecipeParameter('gimasterflat','bsremove-method',group='Bkg Sub',description='Backgruond subtraction method'))
            paramList.append(reflex.RecipeParameter('gimasterflat','bsremove-yorder',group='Bkg Sub',description='Background subtraction yorder of 2-D polynomial fit'))
            paramList.append(reflex.RecipeParameter('gimasterflat','sloc-start',group='LOC',description='Bin along x-axis. [-1]'))
            paramList.append(reflex.RecipeParameter('gimasterflat','sloc-binsize',group='LOC',description='Initial localization detection xbin size. [-1]'))
            paramList.append(reflex.RecipeParameter('gimasterflat','sloc-retries',group='LOC',description='Initial localization detection xbin retries. [10]'))
            paramList.append(reflex.RecipeParameter('gimasterflat','sloc-noise',group='LOC',description='Threshold multiplier. [7.0]'))
            paramList.append(reflex.RecipeParameter('gimasterflat','psf-binsize',group='PSF',description='Size of bin along dispersion axis. [64]'))
            return paramList

        def setWindowHelp(self):
            help_text = """
In this window, the user will interact with the GIRAF flat-field reduction
"""
            return help_text

        def setWindowTitle(self):
            title = 'GIRAF Interactive masterflat reduction'
            return title

except ImportError:
    DataPlotterManager = None


if __name__ == '__main__':
    runInteractiveApp(DataPlotterManager)



# ========================================================================================================
# import the needed modules
try:
    import reflex 
    import sys
    import os
    try :
        from astropy.io import fits as fits
    except:
        import pyfits as fits
    import numpy
    import_sucess = True

    def median_multi( enable=False ):

        # get Kepler parameters:
        parameters= {}
        for p in sys.argv[1:]:
            parameters[p.split("=")[0]]= p.split("=")[1]

        #Define one output file for every input file:
        #outputs.out_sof = inputs.in_sof
        files = inputs.in_sof.files

        scifiles = []
        for file in files :
            if file.category == 'SCIENCE' or file.category == 'STD' :
                scifiles.append(file)

        outlist=list()
        if len(scifiles) > 0 :
            cube=[]
            median_files=[]
            oldpurpose=scifiles[0].purposes[0]
            for ifile,file in enumerate(scifiles):
                # define file names:
                raw_file=file.name
                raw=fits.open(raw_file)
                if enable and ifile > 1:
                    cube.append(raw[0].data)
                    median_image=numpy.median(numpy.array(cube), axis=0)

                    firstraw=fits.open(scifiles[0].name)
                    firstraw[0].data=median_image
                    if 'BZERO' in firstraw[0].header :
                        del firstraw[0].header['BZERO']

                    # save as new file:
                    if ifile == len(scifiles)-1:
                        median_file= parameters["--products-dir"] + "/" + \
	                       (scifiles[0].name).replace(".fits", "_median.fits").split("/")[-1]
                    else:
                        median_file= parameters["--products-dir"] + "/" + \
	                       (scifiles[0].name).replace(".fits", "_median"+str(ifile+1)+".fits").split("/")[-1]
                    median_files.append(median_file)

                    # update header:
                    for i in range(ifile):
                        filename=(scifiles[i].name).split("/")[-1]
                        firstraw[0].header.append(("IFILE"+str(i), filename, "input file for median"))
                    firstraw[0].header.append(("NFILES", ifile+1, "number of input files"))
                    if ifile == len(scifiles)-1:
                        firstraw[0].header.append(("CMETHOD", "MEDIAN", "method for bad pixel cleaning"))
                        firstraw[0].header.append(("HIERARCH ESO PRO CMETHOD", "MEDIAN", "method for bad pixel cleaning"))
                        firstraw[0].header.append(("HIERARCH ESO DRS CMETHOD", "MEDIAN", "method for bad pixel cleaning"))
                    else:
                        firstraw[0].header.append(("CMETHOD", "SUBMED_%d" %(ifile+1), "method for bad pixel cleaning"))
                        firstraw[0].header.append(("HIERARCH ESO PRO CMETHOD", "SUBMED_%d" %(ifile+1), "method for bad pixel cleaning"))
                        firstraw[0].header.append(("HIERARCH ESO DRS CMETHOD", "SUBMED_%d" %(ifile+1), "method for bad pixel cleaning"))

                    # write fits file:
                    try:
                        firstraw.writeto(median_file, output_verify='silentfix+ignore', checksum=True, overwrite=True)
                    except:
                        firstraw.writeto(median_file, output_verify='silentfix+ignore', checksum=True, clobber=True)
                    firstraw.close()

                '''
                # Don't do this now that we have implemented optional keys in the renamed filename
                # And write out a copy of the individual raw files
                raw[0].header.append(("NFILES", 1, "number of input files"))
                raw[0].header.append(("CMETHOD", "NONE", "method for bad pixel cleaning"))
                raw[0].header.append(("HIERARCH ESO PRO CMETHOD", "NONE", "method for bad pixel cleaning"))
                raw[0].header.append(("HIERARCH ESO DRS CMETHOD", "NONE", "method for bad pixel cleaning"))
                try:
                    raw.writeto(parameters["--products-dir"] + "/" + os.path.basename(file.name), overwrite=True)
                except:
                    raw.writeto(parameters["--products-dir"] + "/" + os.path.basename(file.name), clobber=True)
                '''
                # create output SOF:
                # first, lets put in the individual files with a new purpose:
                #outlist.append((reflex.FitsFile(parameters["--products-dir"] + "/" + os.path.basename(file.name),file.category,None,[str("orig/")+oldpurpose])))
                outlist.append((reflex.FitsFile(file.name,file.category,None,[str("orig/")+oldpurpose])))
                raw.close()

            # now add the new median files:
            if enable :
                if False :
                    # Add all the SUBMED versions...
                    for imed, median_file in enumerate(median_files):
                        outlist.append((reflex.FitsFile(median_file,scifiles[0].category,None,[str("med"+str(imed+3)+"/") + oldpurpose])))
                elif len(median_files) > 0 :
                    # Add only the final full median...
                    outlist.append((reflex.FitsFile(median_files[-1],scifiles[0].category,None,[str("med/") + oldpurpose])))

            #for file in outlist:
            #    print("output:", file.name, file.category, file.purposes)

        #send output to EsoReflex:
        newsof = reflex.SetOfFiles(inputs.in_sof.datasetName,outlist)
        outputs.out_sof = newsof


except ImportError:
    import_sucess = False
    print("Error importing one or more of the modules pyfits")

if __name__ == '__main__':

    # Hacked from...
    # example script to process input files and produce one output file for
    # each processed input.
    parser = reflex.ReflexIOParser()

    #Define ports:
    parser.add_option("-i", "--in_sof", dest="in_sof")
    parser.add_option("-e", "--enable", dest="enable", default='false')
    parser.add_output("-o", "--out_sof", dest="out_sof")

    inputs  = parser.get_inputs()
    outputs = parser.get_outputs()
  
    #Check if import failed or not
    if import_sucess :
        median_multi( inputs.enable.lower() == 'true' )
    else :
        outlist=list()
        for file in inputs.in_sof.files:
            if file.category == 'SCIENCE' or file.category == 'STD' :
                outlist.append((reflex.FitsFile(file.name,file.category,file.checksum,file.purposes)))

        outputs.out_sof = reflex.SetOfFiles(inputs.in_sof.datasetName,outlist)

    #send output to EsoReflex:
    parser.write_outputs()

    sys.exit()



# ========================================================================================================
#
# John Pritchard
#
# Developed from esotk_fringe_obj_mask.py by Mark Neeser
#
# A python interactive window for use in the Austrian in-kind
# fringe-correction Reflex workflow.  This interactive window
# shows the results of the fringe-correction pipeline, allowing
# the User to compare the input image with the fringe-corrected
# image, and the fringe map.   From this interactive Actor the
# User can modify pipeline parameters and re-initiate the
# processing.
#
#
#
# ========================================================================================================

import os
import re
import sys
sys.path.append(os.path.dirname(__file__))
from esotk_giraf_gui import BaseDataPlotterManager, runInteractiveApp

try:
    # Import non-standard but needed modules. These must be as subset of what
    # the esotk_fringe_gui module tries to load.
    import reflex
    import pipeline_product
    import pipeline_display
    import matplotlib
    from giraf_plot_common import *

    class DataPlotterManager(BaseDataPlotterManager):

        # This function will read all the columns, images and whatever is needed
        # from the products. The variables , self.plot_x, self.plot_y, etc...
        # are used later in function plotProductsGraphics().
        def readFitsData(self, fitsFiles):
            
            # If fitsFiles doesn't have FF_PSFCENTROID, recipe must have failed, so search for tmp products...
            #Read all the products
            frames = dict()
            for frame in fitsFiles :
                if frame == '' :
                    continue
                category = frame.category
                frames[category] = frame
            self.have_ARC_RBNSTECTRA=True
            if not "ARC_RBNSPECTRA" in frames :
                self.have_ARC_RBNSTECTRA=False
                import glob
                for F in glob.glob('../../giwavecalibration_1/latest/*.fits') :
                    HDUlist=pyfits.open(F)
                    try :
                        fitsFiles.append(reflex.FitsFile(os.path.abspath(F), HDUlist[0].header['HIERARCH ESO PRO CATG'], None, None))
                    except :
                        pass
                    HDUlist.close()

            # Control variable to check if the interesting files were found in
            # the input.
            self.arc_frames = []
            self.science_frames = []
            self.raw_frames = []
            self.frameLabels = {}

            #Initialise the objects to read/display
            self.arc_rbnspectra     = None
            self.arc_raw = None
            self.objid = 1

            # Read all the products
            ff = {}
            for frame in fitsFiles:
                if frame == '':
                    continue
                category = frame.category
                if category == 'ARC_RBNSPECTRA':
                    self.frameLabels[frame.name]=frame.category+': '+frame.name
                    ff[frame.category]=frame
                    self.arc_rbnspectra   = frame
                elif category == 'ARC_SPECTRUM':
                    self.raw_frames.append(frame)
            if self.arc_rbnspectra is not None :
                self.science_frames.append(ff['ARC_RBNSPECTRA'])

            if self.arc_rbnspectra is not None :
                self.arc_rbnspectraPlot   = PlotableARC_RBNSPECTRA(self.arc_rbnspectra)

            self.science_found = len(self.science_frames) > 0
            self.raw_found = len(self.raw_frames) > 0

            self.DPR_CATG = "unknown"
            self.DPR_TYPE = "unknown"
            self.DPR_TECH = "unknown"
            if self.raw_found:
                HDUlist=pyfits.open(self.raw_frames[0].name)
                self.DPR_CATG = HDUlist[0].header['HIERARCH ESO DPR CATG']
                self.DPR_TYPE = HDUlist[0].header['HIERARCH ESO DPR TYPE']
                self.DPR_TECH = HDUlist[0].header['HIERARCH ESO DPR TECH']
                HDUlist.close()

            if self.science_found:
                self.esorex_log_file_name  = re.sub(".*/giraffe/" ,"../../", os.path.dirname(self.science_frames[0].name))+"/log_dir/esorex.log"

        def loadProducts(self, index):
            """
            Loading the plots...
            """
            super(DataPlotterManager, self).loadProducts(index)
            if self.arc_rbnspectra is not None :
                if self.science_frames[index].category == 'FF_PSFCENTROID' :
                    self.arc_rbnspectraPlot   = PlotableARC_RBNSPECTRA(self.arc_rbnspectra)
                
        # This function creates all the subplots. It is responsible for the
        # plotting layouts.
        # There can different layouts, depending on the availability of data
        # Note that subplot(I,J,K) means the Kth plot in a IxJ grid
        # Note also that the last one is actually a box with text, no graphs.
        def addSubplots(self, figure):
            self.list_subplot_image     = []
            if self.arc_rbnspectra is not None :
                gs = matplotlib.gridspec.GridSpec(100,1)
                self.subplot_arc_rbnspectra  = figure.add_subplot(gs[0:92,0])
                self.subplot_txtinfo         = figure.add_subplot(gs[95:100,0])
                self.setupFigure(figure)
            else :
                self.subtext_nodata      = figure.add_subplot(1,1,1)

        # This is the function that makes the plots.
        # Add new plots or delete them using the given scheme.
        # The data has been already stored in self.plot_x, self.plot_xdif, etc ...
        # It is mandatory to add a tooltip variable to each subplot.
        # One might be tempted to merge addSubplots() and plotProductsGraphics().
        # There is a reason not to do it: addSubplots() is called only once at
        # startup, while plotProductsGraphics() is called always there is a
        # resize.
        def plotProductsGraphics(self):

            if self.arc_rbnspectra is not None :

                self.plot_arc_rbnspectra()

                #Additional text info
                self.showTextInfo()

            else :
                #Data not found info
                self.showNoData()

        def plot_arc_rbnspectra(self) :
            title_arc_rbnspectra   = 'Wavelength calibration: Rebinned spectra [ARC_RBNSPECTRA] '
            tooltip_arc_rbnspectra = """Wavelength calibration: rebinned spectra.
Should be smooth."""
            self.arc_rbnspectraPlot.plot(self.subplot_arc_rbnspectra, title_arc_rbnspectra,
                             tooltip_arc_rbnspectra)

        def showTextInfo(self) :
            self.subplot_txtinfo.set_axis_off()
            try:
                mode = self.arc_rbnspectraPlot.rbnspectra.readKeyword('HIERARCH ESO INS SLITS ID')
            except:
                mode = 'unknown'
            try:
                filter_name = self.arc_rbnspectraPlot.rbnspectra.readKeyword('HIERARCH ESO INS FILT NAME')
            except:
                filter_name = 'unknown'
            try:
                grat_wlen = self.arc_rbnspectraPlot.rbnspectra.readKeyword('HIERARCH ESO INS GRAT WLEN')
            except:
                grat_wlen = 'unknown'
            try:
                raw1_name = self.arc_rbnspectraPlot.rbnspectra.readKeyword('HIERARCH ESO PRO REC1 RAW1 NAME')
            except:
                raw1_name = 'unknown'
            self.subplot_txtinfo.text(0.1, 0., 'DPid RAW1: %s\nDPR: %s :: %s :: %s\nMode/Setting/WLen: %s/%s/%s'
                               %(raw1_name,
                                 self.DPR_CATG,self.DPR_TYPE,self.DPR_TECH,
                                 mode,filter_name,str(grat_wlen)
                                ),
                               ha='left', va='top', weight='bold')

        def showNoData(self) :
            #Data not found info
            self.subtext_nodata.set_axis_off()
            self.text_nodata = 'Wave calibration data not found in the products\n(PRO.CATG=ARC_RBNSPECTRA)'
            self.subtext_nodata.text(0.1, 0.6, self.text_nodata, color='#11557c', fontsize=18,
                               ha='left', va='center', alpha=1.0)
            self.subtext_nodata.tooltip='ARC_RBNSPECTRA data not found in the products'


        def plotWidgets(self) :
            widgets = list()
            if False and self.arc_rbnspectra is not None :
                # Clickable subplot
                self.clickablesmapped = InteractiveClickableSubplot(
                       self.subplot_arc_rbnspectra, self.setExtractedObject)
                widgets.append(self.clickablesmapped)
            return widgets

        def setExtractedObject(self, point) :
            obj = self.arc_rbnspectra.getObjectInPosition(point.ydata)
            if obj != -1 :
                self.sci_extracted.selectObject(obj)
            self.plotSciExtracted()

    
        # This function specifies which are the parameters that should be presented
        # in the window to be edited.
        # Note that the parameter has to be also in the in_sop port (otherwise it
        # won't appear in the window)
        # The descriptions are used to show a tooltip. They should match one to one
        # with the parameter list
        # Note also that parameters have to be prefixed by the 'recipe name:'
        def setInteractiveParameters(self):
            paramList = list()
            paramList.append(reflex.RecipeParameter('giwavecalibration','bsremove-method',group='Bkg Sub',description='Backgruond subtraction method'))
            paramList.append(reflex.RecipeParameter('giwavecalibration','bsremove-yorder',group='Bkg Sub',description='Background subtraction yorder of 2-D polynomial fit'))
            paramList.append(reflex.RecipeParameter('giwavecalibration','wcal-lswidth',group='WCAL',description='List of window widths [pxl] used for line detection'))
            #paramList.append(reflex.RecipeParameter('giwavecalibration','wcal-slit',group='WCAL',description='Controls the slit geometry calibration'))
            return paramList

        def setWindowHelp(self):
            help_text = """
In this window, the user will interact with the GIRAF wavelength calibration reduction
"""
            return help_text

        def setWindowTitle(self):
            title = 'GIRAF Interactive wavecalibration reduction'
            return title

except ImportError:
    DataPlotterManager = None


if __name__ == '__main__':
    runInteractiveApp(DataPlotterManager)



# ========================================================================================================
#
# John Pritchard
#
# ========================================================================================================
# import the needed modules
try:
    import reflex 
    import sys
    import os
    try :
        from astropy.io import fits as pyfits
    except:
        import pyfits
    import logging
    import astroscrappy
    import_sucess = True

    def astroscrappy_multi():

        # get Kepler parameters:
        parameters= {}
        for p in sys.argv[1:]:
            parameters[p.split("=")[0]]= p.split("=")[1]

        #Define one output file for every input file:
        outputs.out_sof = inputs.in_sof
        files = outputs.out_sof.files

        masterbiases={}
        badpixmaps={}
        for file in files:
            if file.category == 'MASTER_BIAS' :
                for p in file.purposes :
                    masterbiases[p]=file
            if file.category == 'BAD_PIXEL_MAP' :
                for p in file.purposes :
                    badpixmaps[p]=file

        outlist=list()
        for file in files:
            #select a file category:
            if file.category in ['SCIENCE', 'STD' ] :

                purpose=file.purposes[0]
                try: 
                    HDU=pyfits.open(file.name)
                except:
                    try:
                        logging.error('Could not open %s' %(file.name))
                    except:
                        pass
                    sys.exit(1)
                if purpose in masterbiases :
                    try:
                        bHDU=pyfits.open(masterbiases[purpose].name)
                    except:
                        try:
                            logging.error('Could not open MASTER_BIAS %s' %(masterbiases[purpose].name))
                        except:
                            pass
                        sys.exit(1)

                if not 'CRMASK' in HDU :
                    HDU.append(HDU[0])
                    HDU[-1].header['EXTNAME']='CRMASK'
                    HDU['CRMASK'].header['HIERARCH ESO PRO CATG']='CRMASK'
                    

                if purpose in masterbiases :
                    mask, HDU[0].data = astroscrappy.detect_cosmics(
                        HDU[0].data-bHDU[0].data, 
                        gain=HDU[0].header['HIERARCH ESO DET OUT1 CONAD'], 
                        readnoise=HDU[0].header['HIERARCH ESO DET OUT1 RON']
                      )
                else :
                    mask, HDU[0].data = astroscrappy.detect_cosmics(
                        HDU[0].data,
                        gain=HDU[0].header['HIERARCH ESO DET OUT1 CONAD'], 
                        readnoise=HDU[0].header['HIERARCH ESO DET OUT1 RON']
                      )

                #HDU[0].header['BZERO']=0.
                if 'BZERO' in  HDU[0].header :
                    del HDU[0].header['BZERO']
                HDU[0].data=HDU[0].data/HDU[0].header['HIERARCH ESO DET OUT1 CONAD']
                if purpose in masterbiases :
                    HDU[0].data+=bHDU[0].data
                HDU['CRMASK'].data=mask.astype(int)
                HDU['CRMASK'].header['BZERO']=0.
                HDU[0].header.append(("NFILES", 1, "number of input files"))
                HDU[0].header.append(("CMETHOD", "ASTROSCRAPPY", "method for bad pixel cleaning"))
                HDU[0].header.append(("HIERARCH ESO DRS CMETHOD", "ASTROSCRAPPY", "method for bad pixel cleaning"))
                HDU[0].header.append(('HIERARCH ESO DRS ASCRAP',True,'AstroScrappy Cosmic Ray cleaning applied'))
                if purpose in masterbiases :
                    HDU[0].header.append(('HIERARCH ESO DRS ASCRAP BIASSUB',True,'MASTER_BIAS subtracted?'))
                    HDU[0].header.append(('HIERARCH ESO DRS ASCRAP BIAS NAME',os.path.basename(masterbiases[purpose].name)[0:80-len('HIERARCH ESO ASCRAP BIAS NAME')-5],'MASTER BIAS used'))
                HDU[0].header.append(('HIERARCH ESO DRS ASCRAP RON',HDU[0].header['HIERARCH ESO DET OUT1 RON'],'Readout noise'))
                HDU[0].header.append(('HIERARCH ESO DRS ASCRAP GAIN',HDU[0].header['HIERARCH ESO DET OUT1 CONAD'],'Gain'))

                # define file names:
                clean_file= parameters["--products-dir"] + "/" + \
                       file.name.replace(".fits", "_AS_clean.fits").split("/")[-1]
                file.name=clean_file
                try:
                    HDU.writeto(clean_file, outpur_verify='silentfix+ignore', overwrite=True)
                except:
                    HDU.writeto(clean_file, output_verify='silentfix+ignore', clobber=True)
                HDU.close()
                if purpose in masterbiases :
                    bHDU.close()

                # update purpose
                outlist.append((reflex.FitsFile(file.name,file.category,None,[str("AS_crclean/")+file.purposes[0]])))

        #send output to EsoReflex:
        newsof = reflex.SetOfFiles(inputs.in_sof.datasetName,outlist)
        outputs.out_sof = newsof
        #for file in outlist:
            #print("output:", file.name, file.category, file.purposes)


except ImportError:
    import_sucess = False
    print("Error importing one or more of the modules astropy.io/pyfits, astroscrappy")

except :
    raise

if __name__ == '__main__':
  
    # Hacked from...
    # example script to process input files and produce one output file for
    # each processed input.

    parser = reflex.ReflexIOParser()

    #Define ports:
    parser.add_option("-i", "--in_sof", dest="in_sof")
    #parser.add_option("-e", "--enable", dest="enable", default='false')
    parser.add_output("-o", "--out_sof", dest="out_sof")

    inputs  = parser.get_inputs()
    outputs = parser.get_outputs()

    #Check if import failed or not
    if import_sucess : #and inputs.enable == "true" :
        astroscrappy_multi()
    else :
        # write out an empty sof
        outputs.out_sof = reflex.SetOfFiles(inputs.in_sof.datasetName,list())

    parser.write_outputs()

    sys.exit()



# ========================================================================================================
# import the needed modules
try:
    import reflex
    import sys
    import os
    try :
        from astropy.io import fits as pyfits
        import_astropy_io=True
    except:
        import pyfits
        import_astropy_io=False
    import logging
    import PyCosmic
    import numpy as np
    from scipy import interpolate
    import copy

    import_sucess = True

    def pycosmic_multi(nan_corr_method='1D'):

        # get Kepler parameters:
        parameters= {}
        for p in sys.argv[1:]:
            parameters[p.split("=")[0]]= p.split("=")[1]

        #Define one output file for every input file:
        outputs.out_sof = inputs.in_sof
        files = outputs.out_sof.files

        masterbiases={}
        badpixmaps={}
        for file in files:
            if file.category == 'MASTER_BIAS' :
                for p in file.purposes :
                    masterbiases[p]=file
            if file.category == 'BAD_PIXEL_MAP' :
                for p in file.purposes :
                    badpixmaps[p]=file

        #Define PyCosmic parameters:
        args=lambda: None
        args.rdnoise="HIERARCH ESO DET OUT1 RON"
        args.siglim=5.0
        args.fwhm=1.7 #2.5 # 2.0
        args.rlim=0.5 # 1.2
        args.iter=5 #5
        #args.iter=1 #5
        args.replacebox=[5,5]
        args.radius=0
        args.gain="HIERARCH ESO DET OUT1 CONAD"
        args.parallel=True
        args.verbose=True

        outlist=list()
        for file in files:
            #select a file category:
            if file.category in ['SCIENCE', 'STD' ] :
                purpose=file.purposes[0]
                if purpose in masterbiases :
                    try:
                        bHDU=pyfits.open(masterbiases[purpose].name)
                    except:
                        try:
                            logging.error('Could not open MASTER_BIAS %s' %(masterbiases[0].name))
                        except:
                            pass
                        sys.exit(1)
                if purpose in badpixmaps :
                    try:
                        bpmHDU=pyfits.open(badpixmaps[purpose].name)
                    except:
                        try:
                            logging.error('Could not open BAD_PIXEL_MAP %s' %(badpixmaps[0].name))
                        except:
                            pass
                        sys.exit(1)
                # define file names:
                raw_file=file.name
                bs_file=parameters["--products-dir"] + "/" + \
                       raw_file.replace(".fits", "_bs.fits").split("/")[-1]
                clean_file=parameters["--products-dir"] + "/" + \
                       raw_file.replace(".fits", "_PC_clean.fits").split("/")[-1]
                mask_file=parameters["--products-dir"] + "/" + \
                       raw_file.replace(".fits", "_PC_mask.fits").split("/")[-1]
                file.name=clean_file

                bs=pyfits.open(raw_file)
                if purpose in masterbiases :
                    bs[0].data=bs[0].data-bHDU[0].data
                    #bs[0].header['BZERO']=0.
                    if 'BZERO' in  bs[0].header :
                        del bs[0].header['BZERO']
                if purpose in badpixmaps :
                    bs.append(bpmHDU[0])
                    bs[-1].header['EXTNAME']='BADPIX'

                try:
                    bs.writeto(bs_file, output_verify='silentfix+ignore', overwrite=True)
                except:
                    bs.writeto(bs_file, output_verify='silentfix+ignore', clobber=True)

                # detCos
                #Detects and removes cosmics from astronomical images based on Laplacian edge 
                #detection scheme combined with a PSF convolution approach (Husemann  et al. in prep.).
           
                #IMPORTANT: 
                #The image and the readout noise are assumed to be in units of electrons.
                #The image also needs to be BIAS subtracted!
                #The gain can be entered to convert the image from ADUs to electros, when this is down already set gain=1.0 as the default.
                # clean cosmics:
                PyCosmic.detCos(
                    bs_file,
                    mask_file,
                    clean_file,
                    args.rdnoise,
                    gain=args.gain,
                    sigma_det=args.siglim,
                    rlim=args.rlim,
                    iter=args.iter,
                    fwhm_gauss=args.fwhm,
                    replace_box=args.replacebox,
                    increase_radius=args.radius,
                    parallel=args.parallel,
                    verbose=args.verbose,
                )

                #...the output of PyCosmic is multiplied by the gain, divide it out again...
                clean=pyfits.open(clean_file, 'update')

                if purpose in masterbiases :
                    clean[0].data=clean[0].data/clean[0].header[args.gain]+bHDU[0].data
                else :
                    clean[0].data=clean[0].data/clean[0].header[args.gain]
                # Fix any NaNs...
                number_of_NaNs=0
                NaN_correction_method='None'
                if np.any(np.isnan(clean[0].data)) :
                    _o=pyfits.open(raw_file)
                    _d=np.shape(clean[0].data)
                    _x=np.tile(np.arange(_d[1]),_d[0])
                    _y=np.repeat(np.arange(_d[0]),_d[1])
                    nans, x = nan_helper(np.ravel(clean[0].data))
                    number_of_NaNs=len(_x[nans])
                    NAN_INTERP_METHOD=os.environ.get('GIRAF_PYCOSMIC_NAN_INTERP_METHOD',nan_corr_method).lower()
                    print('Control the NAN_INTERP_METHOD by setting GIRAF_PYCOSMIC_NAN_INTERP_METHOD=["1D", "2D", "orig_pix_value"].')
                    print('Default value is "1D".')
                    print("Note: ENV-VAR method doesn't work in esoreflex environment if esoreflex.inherit-environment=FALSE.")
                    print("Instead set it in the workflow.")
                    if NAN_INTERP_METHOD in ['orig_pix_value',] :
                        ## Replace the NaN with the original pixel value...
                        print('Fixing %d NaNs by replacing with original pixel value...' %(number_of_NaNs))
                        NaN_correction_method='OPIXVAL'
                        for [i,j] in np.transpose(np.reshape(list(_y[nans])+list(_x[nans]),[2,len(_x[nans])])) :
                            clean[0].data[i,j]=_o[0].data[i,j]
                    elif NAN_INTERP_METHOD in ['2d',] :
                        ## 2-D interpolation...
                        print('Fixing %d NaNs with 2-D interpolation...' %(number_of_NaNs))
                        NaN_correction_method='2D-INTERP'
                        for [i,j] in np.transpose(np.reshape(list(_y[nans])+list(_x[nans]),[2,len(_x[nans])])) :
                            _npx=7
                            _npy=7
                            __x=np.tile(np.arange(_npx)-_npx//2,_npy)
                            __y=np.repeat(np.arange(_npy)-_npy//2,_npx)
                            rclean=np.ravel(cd[
                                np.max([0,(i-_npx//2)]):np.min([(i+_npx//2+1),_d[0]-1]),
                                np.max([0,(j-_npy//2)]):np.min([(j+_npy//2+1),_d[1]-1])
                            ])
                            ## Here we assume that the number of NaNs in the 7x7 box is sufficiently small that we can get
                            ## a good interpolation from the remaining NaNs
                            ## The one example looked at so far (GIRAF.2007-04-09T06:52:05.052) had only a single NaN
                            nans, x = nan_helper(rclean)
                            rclean[nans]=0.
                            _f=interpolate.interp2d(__x[x(~nans)],__y[x(~nans)],rclean[x(~nans)], kind='linear')
                            clean[0].data[i,j]=_f(0,0)[0]
                    else :
                        ## 1-D interpolation
                        # Interpolate the values of NaN pixels from the neighbouring pixels of the same column...
                        print('Fixing %d NaNs with 1-D interpolation...' %(number_of_NaNs))
                        NaN_correction_method='1D-INTERP'
                        nans, x = nan_helper(np.sum(clean[0].data,axis=0))
                        for i in x(nans) :
                            nnans, xx = nan_helper(clean[0].data[:,i])
                            clean[0].data[:,i][nnans]= np.interp(
                                xx(nnans), xx(~nnans), clean[0].data[:,i][~nnans]
                            )
                    '''
                    clean.append(pyfits.ImageHDU(header=clean[0].header,data=_o[0].data-clean[0].data))
                    clean[-1].header['HIERARCH ESO PRO CATG']="NAN_DIFF"
                    '''
                # update header
                clean[0].header.append(("HIERARCH ESO PRO CMETHOD", "PYCOSMIC", "method for bad pixel cleaning"))
                clean[0].header.append(("HIERARCH ESO DRS CMETHOD", "PYCOSMIC", "method for bad pixel cleaning"))
                clean[0].header.append(("HIERARCH ESO DRS NANS NUM", number_of_NaNs, "Number of NaNs detected in cleaned frame"))
                clean[0].header.append(("HIERARCH ESO DRS NANS METHOD", NaN_correction_method, "Method used to correct NaNs"))
                clean[0].header.append(("CMETHOD", "PYCOSMIC", "method for bad pixel cleaning"))
                clean[0].header.append(("NFILES", len(files), "number of CR cleaned files"))
                clean[0].header.append(("CRSIGLIM", args.siglim, "PyCosmic signifiance threshold"))
                clean[0].header.append(("CRRLIM", args.rlim, "PyCosmic contrast limit"))
                clean[0].header.append(("CRITER", args.iter, "PyCosmic number of iterations"))
                clean[0].header.append(("CRGFWHM", args.fwhm, "PyCosmic FWHM of convolution kernel"))
                clean[0].header.append(("CRRBOX", str(args.replacebox), "PyCosmic box to computer cleaned value"))
                clean[0].header.append(("CRRADIUS", args.radius, "PyCosmic number of neighboring pixels"))
                clean[0].header.append(('CRRON',clean[0].header['HIERARCH ESO DET OUT1 RON'],'Readout noise'))
                clean[0].header.append(('CRGAIN',clean[0].header['HIERARCH ESO DET OUT1 CONAD'],'Gain'))
                if purpose in masterbiases :
                    clean[0].header.append(('CRMBIAS',os.path.basename(masterbiases[purpose].name)[0:80-len('CRMBIAS')-5],'MASTER_BIAS used'))
                if purpose in badpixmaps :
                    clean[0].header.append(('CRBPM',os.path.basename(badpixmaps[purpose].name)[0:80-len('CRBPM')-5],'BAD_PIX_MAP used'))
                clean[0].header.append(('HIERARCH ESO DRS PYCOSMIC',True,'AstroScrappy Cosmic Ray cleaning applied'))
                if purpose in masterbiases :
                    clean[0].header.append(('HIERARCH ESO DRS PYCOSMIC BIASSUB',True,'MASTER_BIAS subtracted?'))
                    clean[0].header.append(('HIERARCH ESO DRS PYCOSMIC BIAS NAME',os.path.basename(masterbiases[purpose].name)[0:80-len('HIERARCH ESO PYCOSMIC BIAS NAME')-5],'MASTER BIAS used'))
                clean[0].header.append(('HIERARCH ESO DRS PYCOSMIC RON',clean[0].header['HIERARCH ESO DET OUT1 RON'],'Readout noise'))
                clean[0].header.append(('HIERARCH ESO DRS PYCOSMIC GAIN',clean[0].header['HIERARCH ESO DET OUT1 CONAD'],'Gain'))
                clean.flush()
                clean.close()

                # append remaining extensions from raw files unchanged:
                raw=pyfits.open(raw_file)
                for iext in range(1,len(raw)):
                    if import_astropy_io :
                        pyfits.append(clean_file, raw[iext].data, header=raw[iext].header, checksum=True)
                    else :
                        pyfits.append(clean_file, raw[iext].data, header=raw[iext].header)
                raw.close()
                #clean=pyfits.open(clean_file, 'update')
                #print(numpy.min(clean[0].data[:]),numpy.max(clean[0].data[:]))
                #clean.close()

                # update purpose
                outlist.append((reflex.FitsFile(file.name,file.category,None,[str("PC_crclean/")+file.purposes[0]])))

        #send output to EsoReflex:
        newsof = reflex.SetOfFiles(inputs.in_sof.datasetName,outlist)
        outputs.out_sof = newsof
        #parser.write_outputs()
        #for file in outlist:
            #print("output:", file.name, file.category, file.purposes)


except ImportError:
    import_sucess = False
    print("Error importing one or more of the modules astropy.io/pyfits, PyCosmic")

except :
    raise

# ========================================================================================================
## From https://stackoverflow.com/questions/6518811/interpolate-nan-values-in-a-numpy-array
def nan_helper(y):
    """Helper to handle indices and logical indices of NaNs.

    Input:
        - y, 1d numpy array with possible NaNs
    Output:
        - nans, logical indices of NaNs
        - index, a function, with signature indices= index(logical_indices),
          to convert logical indices of NaNs to 'equivalent' indices
    Example:
        >>> # linear interpolation of NaNs
        >>> nans, x= nan_helper(y)
        >>> y[nans]= np.interp(x(nans), x(~nans), y[~nans])
    """

    return np.isnan(y), lambda z: z.nonzero()[0]
# ========================================================================================================

if __name__ == '__main__':
  
    # Hacked from...
    # example script to process input files and produce one output file for
    # each processed input.

    #send output to EsoReflex:
    parser = reflex.ReflexIOParser()
        
    #Define ports:
    parser.add_option("-i", "--in_sof", dest="in_sof")
    parser.add_option("-m", "--nan_corr_method", dest="nan_corr_method", default='1D')
    parser.add_output("-o", "--out_sof", dest="out_sof")

    inputs  = parser.get_inputs()
    outputs = parser.get_outputs()

    #Check if import failed or not
    if import_sucess : #and inputs.enable == "true" :
        pycosmic_multi( inputs.nan_corr_method )
    else :
        # write out an empty sof
        outputs.out_sof = reflex.SetOfFiles(inputs.in_sof.datasetName,list())

    #send output to EsoReflex:
    parser.write_outputs()

    sys.exit()
